@@ -1,12 +1,12 @@
 /***************************************
-  $Header: /home/amb/wwwoffle/RCS/wwwoffle.c 2.24 1999/01/26 17:42:01 amb Exp $
+  $Header: /home/amb/wwwoffle/RCS/wwwoffle.c 2.39 2001/03/31 19:39:47 amb Exp $
 
-  WWWOFFLE - World Wide Web Offline Explorer - Version 2.4b.
+  WWWOFFLE - World Wide Web Offline Explorer - Version 2.6c.
   A user level program to interact with the server.
   ******************/ /******************
   Written by Andrew M. Bishop
 
-  This file Copyright 1996,97,98 Andrew M. Bishop
+  This file Copyright 1996,97,98,99,2000,01 Andrew M. Bishop
   It may be distributed under the GNU Public License, version 2, or
   any higher version.  See section COPYING of the GNU Public license
   for conditions under which this file may be redistributed.
@@ -29,6 +29,10 @@
 #include "sockets.h"
 #include "errors.h"
 
+#ifndef PATH_MAX
+#define PATH_MAX 4096
+#endif
+
 
 static void usage(int verbose);
 
@@ -37,14 +41,23 @@ static void usage(int verbose);
 typedef enum _Action
 {
  None,                          /*+ Undecided. +*/
+
  Online,                        /*+ Tell the server that we are online. +*/
  Autodial,                      /*+ Tell the server that we are in autodial mode. +*/
  Offline,                       /*+ Tell the server that we are offline. +*/
+
  Fetch,                         /*+ Tell the server to fetch the requested pages. +*/
+
  Config,                        /*+ Tell the server to re-read the configuration file. +*/
+
  Purge,                         /*+ Tell the server to purge pages. +*/
+
  Kill,                          /*+ Tell the server to exit. +*/
- Get,                           /*+ Tell the server to get pages. +*/
+
+ Get,                           /*+ Tell the server to GET pages. +*/
+ Post,                          /*+ Tell the server to POST a page. +*/
+ Put,                           /*+ Tell the server to PUT a page. +*/
+
  Output,                        /*+ Get a page from the server and output it. +*/
  OutputWithHeader               /*+ Get a page and the headers from the server and output it. +*/
 }
@@ -55,9 +68,13 @@ static void add_url_list(char **links);
 static void add_url_file(char *url_file);
 
 /*+ The list of URLs or files. +*/
-char **url_file_list=NULL;
+static char **url_file_list=NULL;
+
 /*+ The number of URLs or files. +*/
-int n_url_file_list=0;
+static int n_url_file_list=0;
+
+/*+ A file descriptor for the spool directory; not used anywhere, but required for linking with spool.o. +*/
+int fSpoolDir=-1;
 
 
 /*++++++++++++++++++++++++++++++++++++++
@@ -67,8 +84,10 @@ int n_url_file_list=0;
 int main(int argc, char** argv)
 {
  int i;
- int recursive_depth=0,recursive_mode=0,force=0;
+ int recursive_mode=0,depth=0;
+ int force=0;
  int stylesheets=0,images=0,frames=0,scripts=0,objects=0;
+ char *config_file=NULL;
 
  Action action=None;
 
@@ -78,8 +97,6 @@ int main(int argc, char** argv)
 
  /* Parse the command line options */
 
- ConfigFile=NULL;
-
  if(argc==1)
     usage(0);
 
@@ -106,6 +123,24 @@ int main(int argc, char** argv)
        continue;
       }
 
+    if(!strcmp(argv[i],"-post"))
+      {
+       if(action!=None)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Post;
+       argv[i]=NULL;
+       continue;
+      }
+
+    if(!strcmp(argv[i],"-put"))
+      {
+       if(action!=None)
+         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
+       action=Put;
+       argv[i]=NULL;
+       continue;
+      }
+
     if(!strcmp(argv[i],"-F"))
       {
        if(action!=None && action!=Get)
@@ -160,41 +195,13 @@ int main(int argc, char** argv)
        continue;
       }
 
-    if(!strcmp(argv[i],"-i"))
-      {
-       fprintf(stderr,"wwwoffle: The '-i' option is being replaced by '-gi'.\n");
-
-       if(action!=None && action!=Get)
-         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
-       action=Get;
-       if(images)
-         {fprintf(stderr,"wwwoffle: Only one '-i' argument may be given.\n"); exit(1);}
-       images=1;
-       argv[i]=NULL;
-       continue;
-      }
-
-    if(!strcmp(argv[i],"-f"))
-      {
-       fprintf(stderr,"wwwoffle: The '-f' option is being replaced by '-gf'.\n");
-
-       if(action!=None && action!=Get)
-         {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
-       action=Get;
-       if(frames)
-         {fprintf(stderr,"wwwoffle: Only one '-f' argument may be given.\n"); exit(1);}
-       frames=1;
-       argv[i]=NULL;
-       continue;
-      }
-
     if(!strncmp(argv[i],"-R",2) || !strncmp(argv[i],"-r",2) || !strncmp(argv[i],"-d",2))
       {
        if(action!=None && action!=Get)
          {fprintf(stderr,"wwwoffle: Only one command at a time.\n\n");usage(0);}
        action=Get;
 
-       if(recursive_depth)
+       if(depth)
          {fprintf(stderr,"wwwoffle: Only one '-d', '-r' or '-R' argument may be given.\n"); exit(1);}
 
        if(argv[i][1]=='d')
@@ -205,10 +212,10 @@ int main(int argc, char** argv)
           recursive_mode=3;
 
        if(argv[i][2])
-          recursive_depth=atoi(&argv[i][2]);
+          depth=atoi(&argv[i][2]);
        else
-          recursive_depth=1;
-       if(recursive_depth<=0)
+          depth=1;
+       if(depth<=0)
          {fprintf(stderr,"wwwoffle: The '-%c' argument may only be followed by a positive integer.\n",argv[i][1]); exit(1);}
 
        argv[i]=NULL;
@@ -222,7 +229,7 @@ int main(int argc, char** argv)
        if(++i>=argc)
          {fprintf(stderr,"wwwoffle: The '-p' argument requires a hostname and optionally a port number.\n"); exit(1);}
 
-       if(ConfigFile)
+       if(config_file)
          {fprintf(stderr,"wwwoffle: The '-p' and '-c' options cannot be used together.\n"); exit(1);}
 
        if((colon=strchr(argv[i],':')))
@@ -250,7 +257,7 @@ int main(int argc, char** argv)
        if(host)
          {fprintf(stderr,"wwwoffle: The '-p' and '-c' options cannot be used together.\n"); exit(1);}
 
-       ConfigFile=argv[i];
+       config_file=argv[i];
 
        argv[i-1]=NULL;
        argv[i]=NULL;
@@ -332,12 +339,18 @@ int main(int argc, char** argv)
  if(action==None)
     action=Get;
 
+ if((action==Post || action==Put) && n_url_file_list>1)
+   {
+    fprintf(stderr,"wwwoffle: The -post and -put options only allow one URL.\n\n");
+    usage(0);
+   }
+
  /* Initialise things. */
 
- if(!ConfigFile && !host && (env=getenv("WWWOFFLE_PROXY")))
+ if(!config_file && !host && (env=getenv("WWWOFFLE_PROXY")))
    {
     if(*env=='/')
-       ConfigFile=env;
+       config_file=env;
     else
       {
        char *colon1,*colon2;
@@ -367,31 +380,35 @@ int main(int argc, char** argv)
 
  InitErrorHandler("wwwoffle",0,1);
 
- if(ConfigFile)
+ InitConfigurationFile(config_file);
+
+ if(config_file)
    {
-    if(ReadConfigFile(2))
-       PrintMessage(Fatal,"Error in configuration file '%s'.",ConfigFile);
+    init_buffer(2);
+
+    if(ReadConfigurationFile(2))
+       PrintMessage(Fatal,"Error in configuration file '%s'.",config_file);
 
     host=GetLocalHost(0);
    }
 
  /* The connections to the WWWOFFLE server. */
 
- if(action!=Get && action!=Output && action!=OutputWithHeader)
+ if(action!=Get && action!=Post && action!=Put && action!=Output && action!=OutputWithHeader)
    {
     int socket;
-    char *buffer=NULL;
+    char *line=NULL;
 
-    socket=OpenClientSocket(host?host:"localhost",port?port:WWWOFFLE_Port);
+    socket=OpenClientSocket(host?host:"localhost",port?port:ConfigInteger(WWWOFFLE_Port));
     init_buffer(socket);
 
     if(socket==-1)
-       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:WWWOFFLE_Port);
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:ConfigInteger(WWWOFFLE_Port));
 
     /* Send the message. */
 
-    if(PassWord)
-       write_formatted(socket,"WWWOFFLE PASSWORD %s\r\n",PassWord);
+    if(ConfigString(PassWord))
+       write_formatted(socket,"WWWOFFLE PASSWORD %s\r\n",ConfigString(PassWord));
 
     if(action==Online)
        write_string(socket,"WWWOFFLE ONLINE\r\n");
@@ -410,9 +427,9 @@ int main(int argc, char** argv)
     else
        write_string(socket,"WWWOFFLE BOGUS\r\n");
 
-    while((buffer=read_line(socket,buffer)))
+    while((line=read_line(socket,line)))
       {
-       fputs(buffer,stdout);
+       fputs(line,stdout);
        fflush(stdout);
       }
 
@@ -424,27 +441,57 @@ int main(int argc, char** argv)
  else if(action==Get)
    {
     URL *Url;
+    char *refresh=NULL;
     struct stat buf;
-    char *refresh;
-
-    refresh=CreateRefreshPath(recursive_depth,recursive_mode,force,
-                              stylesheets,images,frames,scripts,objects);
 
     for(i=0;i<n_url_file_list;i++)
        if(strcmp(url_file_list[i],"-") && stat(url_file_list[i],&buf))
          {
           int socket;
-          char buffer256[257];
+          char buffer[READ_BUFFER_SIZE];
 
-          socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port);
+          socket=OpenClientSocket(host?host:"localhost",port?port:ConfigInteger(HTTP_Port));
           init_buffer(socket);
 
           if(socket==-1)
-             PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:HTTP_Port);
+             PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:ConfigInteger(HTTP_Port));
 
           Url=SplitURL(url_file_list[i]);
 
-          printf("Getting: %s\n",Url->name);
+          if(recursive_mode || depth || force || stylesheets || images || frames || scripts || objects)
+            {
+             char *limit="";
+
+             if(recursive_mode==3)
+                limit="";
+             else if(recursive_mode==2)
+               {
+                limit=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+5);
+                sprintf(limit,"%s://%s/",Url->proto,Url->host);
+               }
+             else if(recursive_mode==1)
+               {
+                char *p;
+                limit=(char*)malloc(strlen(Url->proto)+strlen(Url->host)+strlen(Url->path)+5);
+                sprintf(limit,"%s://%s%s",Url->proto,Url->host,Url->path);
+                p=limit+strlen(limit)-1;
+                while(p>limit && *p!='/')
+                   *p--=0;
+               }
+             else
+                depth=0;
+
+             refresh=CreateRefreshPath(Url,limit,depth,
+                                       force,
+                                       stylesheets,images,frames,scripts,objects);
+
+             if(*limit)
+                free(limit);
+
+             printf("Getting: %s (with recursive options).\n",Url->name);
+            }
+          else
+             printf("Getting: %s\n",Url->name);
 
           if(Url->user)
             {
@@ -453,28 +500,46 @@ int main(int argc, char** argv)
              sprintf(userpass1,"%s:%s",Url->user,Url->pass?Url->pass:"");
              userpass2=Base64Encode(userpass1,strlen(userpass1));
 
-             write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
-                                    "Pragma: wwwoffle\r\n"
-                                    "Accept: */*\r\n"
-                                    "Authorization: Basic %s\r\n"
-                                    "\r\n",
-                             refresh,Url->name,userpass2);
+             if(refresh)
+                write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                                       "Authorization: Basic %s\r\n"
+                                       "Pragma: wwwoffle\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                refresh,userpass2);
+             else
+                write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                                       "Authorization: Basic %s\r\n"
+                                       "Pragma: wwwoffle\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                Url->name,userpass2);
 
              free(userpass1);
              free(userpass2);
             }
           else
-             write_formatted(socket,"GET %s/?%s HTTP/1.0\r\n"
-                                    "Pragma: wwwoffle\r\n"
-                                    "Accept: */*\r\n"
-                                    "\r\n",
-                             refresh,Url->name);
+            {
+             if(refresh)
+                write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                                       "Pragma: wwwoffle\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                refresh);
+             else
+                write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                                       "Pragma: wwwoffle\r\n"
+                                       "Accept: */*\r\n"
+                                       "\r\n",
+                                Url->name);
+            }
 
-          while(read_data(socket,buffer256,256)>0)
+          while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
              ;
 
           CloseSocket(socket);
 
+          free(refresh);
           FreeURL(Url);
          }
        else if(!strcmp(url_file_list[i],"-") || S_ISREG(buf.st_mode))
@@ -486,6 +551,8 @@ int main(int argc, char** argv)
           if(strcmp(url_file_list[i],"-"))
             {
              file=open(url_file_list[i],O_RDONLY);
+             init_buffer(file);
+
              if(file==-1)
                {PrintMessage(Warning,"Cannot open file '%s' for reading.",url_file_list[i]);continue;}
 
@@ -494,20 +561,19 @@ int main(int argc, char** argv)
           else
             {
              file=fileno(stdin);
+             init_buffer(file);
 
              printf("Reading: <stdin>\n");
             }
 
-          init_buffer(file);
-
           strcpy(buffer,"file://localhost");
           if(*url_file_list[i]=='/')
              strcat(buffer,url_file_list[i]);
           else
             {
-             char cwd[256];
+             char cwd[PATH_MAX+1];
 
-             if(!getcwd(cwd,255))
+             if(!getcwd(cwd,PATH_MAX))
              cwd[0]=0;
 
              strcat(buffer,cwd);
@@ -540,27 +606,120 @@ int main(int argc, char** argv)
              add_url_list(links);
 
           FreeURL(Url);
+          free(buffer);
           if(file!=0)
              close(file);
          }
        else
           PrintMessage(Warning,"The file '%s' is not a regular file.",url_file_list[i]);
    }
+ else if(action==Post || action==Put)
+   {
+    URL *Url;
+    int socket,n,length=0;
+    char buffer[READ_BUFFER_SIZE];
+    char *data=(char*)malloc(READ_BUFFER_SIZE+1);
+
+    socket=OpenClientSocket(host?host:"localhost",port?port:ConfigInteger(HTTP_Port));
+    init_buffer(socket);
+
+    if(socket==-1)
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:ConfigInteger(HTTP_Port));
+
+    Url=SplitURL(url_file_list[0]);
+
+    if(action==Post)
+       printf("Posting: %s\n",Url->name);
+    else
+       printf("Putting: %s\n",Url->name);
+
+    while((n=read_data(0,data+length,READ_BUFFER_SIZE))>0)
+      {
+       length+=n;
+       data=(char*)realloc((void*)data,length+READ_BUFFER_SIZE+1);
+      }
+
+    if(action==Post)
+      {
+       data[length]=0;
+       data=URLRecodeFormArgs(data);
+       length=strlen(data);
+      }
+
+    if(Url->user)
+      {
+       char *userpass1=(char*)(malloc(strlen(Url->user)+(Url->pass?strlen(Url->pass):1)+3)),*userpass2;
+
+       sprintf(userpass1,"%s:%s",Url->user,Url->pass?Url->pass:"");
+       userpass2=Base64Encode(userpass1,strlen(userpass1));
+
+       if(action==Post)
+          write_formatted(socket,"POST %s HTTP/1.0\r\n"
+                                 "Authorization: Basic %s\r\n"
+                                 "Content-Type: application/x-www-form-urlencoded\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Pragma: wwwoffle\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,userpass2,length);
+       else
+          write_formatted(socket,"PUT %s HTTP/1.0\r\n"
+                                 "Authorization: Basic %s\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Pragma: wwwoffle\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,userpass2,length);
+
+       free(userpass1);
+       free(userpass2);
+      }
+    else
+      {
+       if(action==Post)
+          write_formatted(socket,"POST %s HTTP/1.0\r\n"
+                                 "Content-Type: application/x-www-form-urlencoded\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Pragma: wwwoffle\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,length);
+       else
+          write_formatted(socket,"PUT %s HTTP/1.0\r\n"
+                                 "Content-Length: %d\r\n"
+                                 "Pragma: wwwoffle\r\n"
+                                 "Accept: */*\r\n"
+                                 "\r\n",
+                          Url->name,length);
+      }
+
+    write_data(socket,data,length);
+
+    write_data(socket,"\r\n",2);
+
+    while(read_data(socket,buffer,READ_BUFFER_SIZE)>0)
+       ;
+
+    CloseSocket(socket);
+
+    FreeURL(Url);
+    free(data);
+   }
  else /* action==Output or action==OutputWithHeader */
    {
     URL *Url;
     int socket;
-    char *buffer=NULL,buffer256[257];
+    char *line=NULL,buffer[READ_BUFFER_SIZE];
     int nbytes;
 
     if(!n_url_file_list)
        PrintMessage(Fatal,"No URL specified to output.");
 
-    socket=OpenClientSocket(host?host:"localhost",port?port:HTTP_Port);
+    socket=OpenClientSocket(host?host:"localhost",port?port:ConfigInteger(HTTP_Port));
     init_buffer(socket);
 
     if(socket==-1)
-       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:HTTP_Port);
+       PrintMessage(Fatal,"Cannot open connection to wwwoffle server %s port %d.",host?host:"localhost",port?port:ConfigInteger(HTTP_Port));
 
     Url=SplitURL(url_file_list[0]);
 
@@ -575,9 +734,9 @@ int main(int argc, char** argv)
        userpass2=Base64Encode(userpass1,strlen(userpass1));
 
        write_formatted(socket,"GET %s HTTP/1.0\r\n"
+                              "Authorization: Basic %s\r\n"
                               "Pragma: wwwoffle\r\n"
                               "Accept: */*\r\n"
-                              "Authorization: Basic %s\r\n"
                               "\r\n",
                        Url->name,userpass2);
 
@@ -591,14 +750,14 @@ int main(int argc, char** argv)
                               "\r\n",
                        Url->name);
 
-    if((buffer=read_line(socket,buffer)))
+    if((line=read_line(socket,line)))
       {
        char *willget="HTTP/1.0 404 WWWOFFLE Will Get\r\n";
        int status;
 
-       sscanf(buffer,"%*s %d",&status);
+       sscanf(line,"%*s %d",&status);
 
-       if(!strcmp(willget,buffer))
+       if(!strcmp(willget,line))
           fprintf(stderr,"The URL is not in the cache but has been requested.\n");
        else if((status>=300 && status<400) && action!=OutputWithHeader)
           fprintf(stderr,"The URL has been moved, check with a browser.\n");
@@ -607,18 +766,18 @@ int main(int argc, char** argv)
        else
          {
           if(action==OutputWithHeader)
-             fputs(buffer,stdout);
+             fputs(line,stdout);
 
-          while((buffer=read_line(socket,buffer)))
+          while((line=read_line(socket,line)))
             {
              if(action==OutputWithHeader)
-                fputs(buffer,stdout);
-             if(buffer[0]=='\r' || buffer[0]=='\n')
+                fputs(line,stdout);
+             if(line[0]=='\r' || line[0]=='\n')
                 break;
             }
 
-          while((nbytes=read_data(socket,buffer256,256))>0)
-             fwrite(buffer256,1,nbytes,stdout);
+          while((nbytes=read_data(socket,buffer,READ_BUFFER_SIZE))>0)
+             fwrite(buffer,1,nbytes,stdout);
          }
       }
     else
@@ -650,8 +809,8 @@ static void usage(int verbose)
 
  if(verbose)
     fprintf(stderr,
-            "(c) Andrew M. Bishop 1996,97,98,99 [       amb@gedanken.demon.co.uk ]\n"
-            "                                   [http://www.gedanken.demon.co.uk/]\n"
+            "(c) Andrew M. Bishop 1996,97,98,99,2000 [       amb@gedanken.demon.co.uk ]\n"
+            "                                        [http://www.gedanken.demon.co.uk/]\n"
             "\n");
 
  fprintf(stderr,
@@ -661,6 +820,7 @@ static void usage(int verbose)
          "       wwwoffle [-o|-O] <url>\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] <url> ...\n"
          "       wwwoffle [-g[Sisfo]] [-F] [-(d|r|R)[<depth>]] [<file>|-] ...\n"
+         "       wwwoffle -post <url> | -put <url>\n"
          "\n"
          "Any of these can also take:  [-p <host>[:<port>] | -c <config-file>]\n"
          "The environment variable WWWOFFLE_PROXY can be set instead of -p or -c options.\n"
@@ -704,6 +864,12 @@ static void usage(int verbose)
             "                        (-r => URLs on the same host)\n"
             "                        (-R => URLs on any host)\n"
             "\n"
+            "wwwoffle -post <url> : Create a request using the POST method, the data is read\n"
+            "                       from stdin and appended to the request.  The user should\n"
+            "                       ensure that the data is correctly url-encoded.\n"
+            "wwwoffle -put <url>  : Create a request using the PUT method, the data is read\n"
+            "                       from stdin and appended to the request.\n"
+            "\n"
             " -p <host>[:<port>]  : The host name and port number to talk to the demon on.\n"
             "                       (Defaults to localhost for the server and\n"
             "                        %d for control port, %d for http proxy port).\n"
